import types

from . import debug, text

pronoun_referents = {
  'last_time': None,
  'last_place': None,
  'last_direction': None,
  'last_person': None,
  'last_group': None,
  'last_male': None,
  'last_female': None,
  'last_thing': None,
  'last_object': None,
  'last_event': None,
  'last_occurrence': None,
  'last_condition': None,
  'last_action': None,
}

def find_matches(tokens, typ, text, **extras):
  '''
  Takes a token list, a type, and a TEXT, and returns pairs of object-set,
  remainder for each matching name. Extra items that map to non-set values are
  ignored.
  '''
  matches = []
  extras = {k: v for k, v in extras.items() if isinstance(v, set)}
  debug.debug("Understanding", "Matching against tokens: {}.".format(tokens))
  debug.debug("Understanding", "  ...with extras: {}.".format(extras))
  for name, remainder in ((' '.join(tokens[:i]), tokens[i:])
                          for i in range(1, len(tokens) + 1)):
    debug.debug("Understanding", "Trying to match name '{}'.".format(name))
    results = text.lookup_name(name) or set()
    if name in extras:
      results = results.union(extras[name])
    results.discard(None)
    for r in results:
      if isinstance(r, typ):
        matches.append((r, remainder))
  debug.debug("Understanding", "Matches: {}.".format(matches))
  return matches

class Matcher:
  '''
  A Matcher is an object that matches against some or all of some user input.
  It defines a match method that takes a sequence of tokens and returns a
  remainder sequence, setting the matched field to either True or False and
  in some cases, setting the bindings field appropriately.
  '''
  name = None
  optional = False
  value = None
  typ = None
  policy="longest"
  matched = False
  bindings = None

  def __init__(
    self,
    name=None,
    optional=False,
    value=None,
    typ=str,
    policy="longest",
  ):
    self.name = name
    self.optional = optional
    self.value = value
    self.policy = policy

    self.matched = False
    self.bindings = {}

    if type(typ) == str:
      if hasattr(text, typ):
        self.typ = getattr(text, typ)
      elif typ in globals():
        self.typ = globals()[typ]
      elif isinstance(__builtins__, types.ModuleType)\
       and hasattr(__builtins__, typ):
        self.typ = getattr(__builtins__, typ)
      elif isinstance(__builtins__, dict) and typ in __builtins__:
        self.typ = __builtins__[typ]
      else:
        raise TypeError("Couldn't find type '{}'!".format(typ))
    elif type(typ) == type:
      self.typ = typ
    else:
      raise TypeError(
        "Bad type for type parameter typ: '{}'".format(type(typ))
      )

class StringMatcher(Matcher):
  '''
  A Matcher that matches a raw string from the input, setting matched as either
  True or False.
  '''
  def __init__(self, optional=False, value=None):
    super().__init__(optional=optional, value=value, typ=str, policy="first")

  def match(self, context, tokens):
    '''
    Matches a string with the exact given value. Does not bind anything.
    '''
    self.matched = False
    if not tokens:
      return tokens
    if ' '.join(tokens[:self.value.count(' ') + 1]) == self.value:
      self.matched = True
      return tokens[1:]
    return tokens

class ObjectMatcher(Matcher):
  '''
  An ObjectMatcher searches for (or constructs) the given type of object and
  binds that object to the given name in the bindings dictionary.
  '''
  def match(self, context, tokens):
    '''
    Calls the find_matches function to find string matches against the given
    tokens among objects of the appropriate type. Sets the matched field to the
    matching object or string and returns the remaining tokens list.
    '''
    global pronoun_referents
    state, here = context
    self.matched = False
    self.bindings = {}
    if not tokens:
      return tokens
    now = state.time
    txt = state.text
    prev = None
    next = None
    times = sorted(txt.times.values(), key=lambda t: t.order)
    idx = times.index(now)
    if idx > 0:
      prev = {times[idx - 1]}
    if idx < len(times) - 1:
      next = {times[idx + 1]}
    if issubclass(self.typ, str):
      results = [(tokens[0], tokens[1:])]
    elif issubclass(self.typ, text.Time):
      results = find_matches(
        tokens,
        self.typ,
        txt,
        previous=prev,
        backward=prev,
        backwards=prev,
        next=next,
        forward=next,
        forwards=next,
        now={now},
        then={pronoun_referents['last_time']},
      )
    elif issubclass(self.typ, text.Direction):
      back = None
      if (pronoun_referents['last_direction']):
        back={pronoun_referents['last_direction'].opposite}
      results = find_matches(
        tokens,
        self.typ,
        txt,
        straight={pronoun_referents['last_direction']},
        back=back,
      )
    elif issubclass(self.typ, text.Place):
      results = find_matches(
        tokens,
        self.typ,
        txt,
        here={here},
        there={pronoun_referents['last_place']},
      )
    elif issubclass(self.typ, text.Thing):
      results = find_matches(
        tokens,
        self.typ,
        txt,
        now={now},
        he={pronoun_referents['last_male']},
        she={pronoun_referents['last_female']},
        em={pronoun_referents['last_person']},
        them={pronoun_referents['last_group']},
        it={pronoun_referents['last_object']},
      )
    elif issubclass(self.typ, text.Event):
      results = find_matches(
        tokens,
        self.typ,
        txt,
        now={now},
        it={pronoun_referents['last_event']},
      )
    elif issubclass(self.typ, text.Condition):
      results = find_matches(
        tokens,
        self.typ,
        txt,
        now={now},
        it={pronoun_referents['last_condition']},
      )
    elif issubclass(self.typ, Action):
      results = find_matches(
        tokens,
        self.typ,
        txt,
        now={now},
        it={pronoun_referents['last_action']},
        last={pronoun_referents['last_action']},
      )
    else:
      raise TypeError("Unknown variable type '{}'".format(self.typ))

    filtered = list(filter(
      lambda m: (not self.value) or (m[0] == self.value),
      results,
    ))

    if not filtered:
      return tokens

    # TODO: Allow for disambiguation!

    if self.policy == "longest":
      result = sorted(filtered, key=lambda match: len(match[1]))[0]
    elif self.policy == "shortest":
      result = sorted(filtered, key=lambda match: -len(match[1]))[0]
    elif self.policy == "first":
      result = filtered[0]
    else:
      raise ValueError("Bad policy: '{}'".format(self.policy))
    # TODO: More policies?
    self.matched = True
    bound, leftovers = result
    self.bindings[self.name] = bound
    if isinstance(bound, text.Time):
      pronoun_referents['last_time'] = bound
    elif isinstance(bound, text.Place):
      pronoun_referents['last_place'] = bound
    elif isinstance(bound, text.Direction):
      pronoun_referents['last_direction'] = bound
    elif isinstance(bound, text.Thing):
      pronoun_referents['last_thing'] = bound
      if isinstance(bound, text.Entity):
        if bound.plural:
          pronoun_referents['last_group'] = bound
        else:
          if bound.gender in ("neutral", "masculine", "feminine"):
            pronoun_referents['last_person'] = bound
          else:
            pronoun_referents['last_object'] = bound
        if bound.gender == "masculine":
          pronoun_referents['last_male'] = bound
        if bound.gender == "feminine":
          pronoun_referents['last_female'] = bound
      elif isinstance(bound, text.Object):
        pronoun_referents['last_object'] = bound
    elif isinstance(bound, text.Event):
      pronoun_referents['last_event'] = bound
    elif isinstance(bound, text.Occurrence):
      pronoun_referents['last_occurrence'] = bound
    elif isinstance(bound, text.Condition):
      pronoun_referents['last_condition'] = bound
    elif isinstance(bound, Action):
      pronoun_referents['last_action'] = bound
    return leftovers

class SequenceMatcher(Matcher):
  '''
  A SequenceMatcher is a matcher that runs a bunch of other matchers
  sequentially and unifies the results.
  '''
  matchers = None
  optional = False
  matched = False
  score = None

  def __init__(self, *matchers, optional=False):
    self.matchers = matchers
    self.optional = optional
    self.matched = False
    self.bindings = {}
    self.score = 0

  def match(self, context, tokens):
    '''
    Unifies bindings for matched sub-matchers, but of course only if they all
    match (or are optional).
    '''
    self.bindings = {}
    self.matched = False
    self.score = 0
    tks = tokens[:]
    for m in self.matchers:
      tks = m.match(context, tks)
      if m.matched:
        if hasattr(m, 'score'):
          self.score += m.score
        else:
          self.score += 1
        self.bindings.update(m.bindings)
      elif not m.optional:
        self.bindings = {}
        return tokens
    self.matched = True
    return tks

class ChoiceMatcher(Matcher):
  '''
  A ChoiceMatcher is a matcher that runs other matchers as alternatives and
  unifies the results.
  '''
  matchers = None
  optional = False
  matched = False
  policy = "first"

  def __init__(self, *matchers, optional=False, policy="first"):
    self.matchers = matchers
    self.optional = optional
    self.policy = policy
    self.matched = False
    self.bindings = {}

  def match(self, context, tokens):
    '''
    Picks a match from the set of successful matches based on the policy,
    and unifies bindings from that. Other matches are disregarded.
    '''
    self.bindings = {}
    self.matched = False
    matches = []
    for m in self.matchers:
      tks = m.match(context, tokens)
      if m.matched:
        matches.append((m, tks))
    if matches:
      self.matched = True
      if self.policy == "first":
        self.bindings.update(matches[0][0].bindings)
        return matches[0][1]
      elif self.policy == "longest":
        longest = sorted(matches, key=lambda m: len(m[1]))[0]
        self.bindings.update(longest[0].bindings)
        return longest[1]
      else:
        raise ValueError("Bad policy: '{}'".format(self.policy))

def build_matcher(template):
  '''
  Builds a single matcher from part of a template.
  '''
  optional = False
  if template[-1] == '?':
    optional = True
    template = template[:-1]
  if '_' in template:
    return SequenceMatcher(
      *[build_matcher(st) for st in template.split('_')],
      optional=optional
    )
  if '|' in template:
    return ChoiceMatcher(
      *[build_matcher(st) for st in template.split('|')],
      optional=optional
    )
  if template[0] == '<' and template[-1] == '>':
    template = template[1:-1]
    if '/' in template:
      typ, name = template.split('/')
    else:
      typ = str
      name = template
    return ObjectMatcher(
      name = name,
      optional = optional,
      value = None,
      typ = typ,
    )
  else:
    return StringMatcher(
      value = template,
      optional = optional,
    )

class Action:
  '''
  An Action is something that the user can do. It includes a regular expression
  for recognizing user commands and some code for carrying out the action.
  '''
  def __init__(self, id=None, template=None, code="", help=""):
    self.id = id
    self.name = id # TODO: Is this alright?
    self.template = template
    self.code = code
    self.help = help

    self.score = 0
    self.matcher = SequenceMatcher(
      *[build_matcher(t) for t in self.template.split()]
    )

    self.vars = []
    tosearch = [self.matcher]
    while tosearch:
      this = tosearch[0]
      tosearch = tosearch[1:]
      if hasattr(this, 'matchers'):
        tosearch.extend(this.matchers)
      if isinstance(this, ObjectMatcher):
        self.vars.append(this.name)

    debug.debug("Construction", "Found vars for action:\n{}".format(self.vars))

  def handle(self, input, context, globals, locals):
    '''
    Handles the given input string, executing the code for this action and
    returning True. If the input string doesn't match this action, however, it
    just returns False, but it also sets its score field to the number of
    tokens matched. When executing the code, this method uses the given globals
    and locals dictionaries, as well as the 'TEXT' variable from the locals
    dictionary. It also uses (and then sets) the various pronoun-resolution
    keys, like last_object and last_person.
    '''
    global pronoun_referents
    leftovers = self.matcher.match(context, input.split(' '))
    self.score = self.matcher.score
    # TODO: More informative error messages!
    if leftovers or not self.matcher.matched:
      debug.debug(
        "Actions",
        "Action match failure",
        "[{}] against [{}]".format(input, self.template)
      )
      return False
    vars = {"input": input}
    vars.update({v:None for v in self.vars})
    vars.update(self.matcher.bindings)
    vars.update(locals)
    vars.update(pronoun_referents)
    exec(self.code, globals, vars)
    return True
