
class multidict:
  '''
  A modified dictionary that contains a value-set for each key, instead of just
  a single value. It is somewhat interoperable with normal dictionaries.
  '''
  def __init__(self, dict=None, **kwargs):
    dict = dict or {}
    dict.update(kwargs)
    self._map = {}
    self.update(dict)

  def __str__(self):
    return str(self._map)

  def __repr__(self):
    return "multidict(" + repr(self._map) + ")"

  def __contains__(self, key):
    return key in self._map

  def __len__(self):
    return len(self._map)

  def __iter__(self):
    return iter(self._map.keys())

  def __getitem__(self, key):
    if key in self._map:
      return self._map[key]
    else:
      raise KeyError("Key '{}' not in multidict.".format(key))

  def __setitem__(self, key, value):
    if isinstance(value, set):
      self._map[key] = value
    else:
      raise TypeError(
        "Raw multidict values must be sets (use map to establish individual "
        "mappings)."
      )

  def __delitem__(self, key):
    del self._map[key]

  def maps_to(self, key, value):
    '''
    Returns True if the given key maps to the given value.
    '''
    if key in self._map:
      return (value in self._map[key])
    else:
      raise KeyError("Key '{}' not in multidict.".format(key))

  def map(self, key, value):
    '''
    Maps the given key to the given value.
    '''
    try:
      hash(value)
    except:
      raise TypeError("A multidict may only contain hashable values!")
    if key in self._map:
      self._map[key] = self._map[key].union({value})
    else:
      self._map[key] = {value}

  def get(self, key):
    '''
    Returns an arbitrary mapping for the given key.
    '''
    if key in self._map:
      return list(self._map[key])[0]
    else:
      raise KeyError("Key '{}' not in multidict.".format(key))

  def map_all(self, key, *values):
    '''
    Maps the given key to all of the given values.
    '''
    for v in values:
      self.map(key, v)

  def unmap(self, key, value):
    '''
    Removes the mapping from key to value (without affecting other mappings of
    the key).
    '''
    if key in self._map:
      if value in self._map[key]:
        self._map[key].discard(value)
        if not self._map[key]:
          del self._map[key]
      else:
        raise KeyError(
          "Key '{}' doesn't map to value '{}' in multidict.".format(
            key,
            value,
          )
        )
    else:
      raise KeyError("Key '{}' not in multidict.".format(key))

  def unmap_all(self, value):
    '''
    Unmaps all mappings for the given value.
    '''
    for key in list(self._map.keys()):
      if value in self._map[key]:
        self._map[key].discard(value)
        if not self._map[key]:
          del self._map[key]

  def update(self, source=None, **kwargs):
    source = source or {}
    source.update(kwargs)
    if isinstance(source, multidict):
      for k, v in source:
        if k in self._map:
          self._map[k] = self._map[k].union(v)
        else:
          self._map[k] = v
    else:
      for k, v in source:
        self.map(k, v)

  def also(self, source=None, **kwargs):
    '''
    Returns a new dictionary that is the result of copying this one and
    updating it with the given source and/or keyword arguments.
    '''
    result = multidict(self)
    result.update(source, **kwargs)
    return result

  def keys(self):
    return self._map.keys()

  def values(self):
    result = set()
    for v in self._map.values():
      result = result.union(v)
    return result

  def items(self):
    return self._map.items()

  def mappings(self):
    '''
    Yields each mapping individually as a twople.
    '''
    for k in self._map:
      for v in self._map[k]:
        yield (k, v)
