import collections, copy, re, sys

from . import debug, formatter, multidict

PROPERTY_ACCESSOR = "has_property_"

def first(thing):
  for it in thing:
    return it

def output(*args, end='\n'):
  print(' '.join(str(a) for a in args), end=end)

def subids(code, idlist, env={}):
  '''
  Takes a piece of code that's going to be executed or evaluated and replaces
  any IDs with appropriate lookups.
  '''
  for id in idlist:
    if id in env:
      debug.debug(
        "Conditions",
        "Warning: id '{}' collided with a built-in value!".format(id)
      )
    if id in code:
      code = re.sub(
        r'\b(' + re.escape(id) + r')\b',
        r'Unique.lookup("""\1""")',
        code
      )
  return code

utils = [
  first,
  output,
]

def LIB(state, place):
  '''
  Takes a state and place and returns the standard library of functions
  relative to that state and place.
  '''
  lib = {
    'Unique': Unique,
    'TEXT': state.text,
    'now': state.time,
    'here': place,
    'state': state,
    'location_of': lambda t: state.locations[t],
    'contains': lambda p, t: state.locations[t] == p,
    'is_present': lambda t: state.locations[t] == place,
    'happened': lambda e: any(any(occ.event == e
                                  for occ in state.trace[t])
                              for t in state.trace),
    'happened_at': lambda e, port:
      (type(port) == Place and any(any(occ.event == e and occ.place == p
                                       for occ in state.trace[t])
                                   for t in state.trace)) or
      (type(port) == Time and any(any(occ.event == e and occ.place == p
                                      for occ in state.trace[t])
                                  for t in state.trace)),
    'a': Typer(),
  }
  times = sorted(state.text.times.values(), key=lambda t: t.order)
  idx = times.index(state.time)
  if idx > 0:
    lib['previous'] = times[idx - 1]
    lib['backward'] = times[idx - 1]
    lib['backwards'] = times[idx - 1]
  if idx < len(times) - 1:
    lib['next'] = times[idx + 1]
    lib['forward'] = times[idx + 1]
    lib['forwards'] = times[idx + 1]
  lib.update({f.__code__.co_name: f for f in utils})
  lib.update({f.__code__.co_name: f for f in formatter.utils})
  return lib

class Typer:
  '''
  This class is like an Articulator (see formatter), but more general. It
  transforms attribute lookups into type lookups. Note that it only looks up
  native types, so it can't resolve Action.
  '''
  def __getattr__(self, attr):
    if attr in globals():
      result = globals()[attr]
      if type(result) == type:
        return result
    return "ERROR: Type '{}' does not exist.".format(attr)

class Unique:
  '''
  A class that enforces a unique 'id' field.
  '''
  ids = {}
  def __init__(self, id=None):
    if id == "":
      raise ValueError("Sorry, but the empty string is not a valid ID.")
    self.id = id
    Unique.ids[id] = self

  @staticmethod
  def lookup(id):
    '''
    Returns the object corresponding to the given id, or None if there is no
    corresponding object.
    '''
    if id in Unique.ids:
      return Unique.ids[id]
    return None

  def __eq__(self, other):
    if isinstance(other, Unique):
      return self.id == other.id
    return False

  def __ne__(self, other):
    return not self == other

  def __hash__(self):
    h = hash(self.id)
    return h + 7 + h % 753

  def __str__(self):
    return "a unique object with ID '{}'".format(self.id)

  def __repr__(self):
    return "<" + str(self) + ">"

  def get_defaults(self):
    '''
    The recursion base-case for the get_defaults method of Anythings. See that
    get_defaults method for more info.
    '''
    return {}

class Anything(Unique):
  '''
  A class that packs keyword constructor arguments into itself, using
  self.defaults to fill in missing arguments. Also inherits from Unique.
  '''
  defaults = {
    'article': "a ", # TODO: Get rid of this?
    'determiner': "the ", # TODO: Get rid of this?
    'gender': "neuter",
    'preposition': "at ",
    'generic': False,
    'properties': set(),
  }
  def __init__(self, **kwargs):
    '''
    An all-purpose constructor that takes keyword arguments, augments them with
    default values defined by the class, and packs them into the resulting
    instance. If no 'id' field is given, it uses the 'name' field if present,
    or constructs an ID if 'idfmt' and/or 'idargs' are given. If 'idfmt' is
    given, this is used as the ID. If 'idargs' is also given, it should be a
    sequence of strings. Each of these is looked up in the combined
    arguments/defaults dictionary, and the values are passed to format() called
    on the ID string that came from 'idfmt' (in the absence of 'idfmt',
    'idargs' is not used). Note that 'name', 'idargs', and 'idfmt' are place
    onto the instance just like all the other keyword arguments, as well as
    being used to construct an ID if necessary.
    '''
    defaults = self.get_defaults()
    defaults.update(kwargs)
    if 'id' not in defaults:
      if 'idfmt' in defaults:
        defaults['id'] = defaults['idfmt']
        if 'idargs' in defaults:
          defaults['id'] = defaults['idfmt'].format(
            *(defaults[val] for val in defaults['idargs'])
          )
      elif 'name' in defaults:
        defaults['id'] = defaults['name']
    if 'name' not in defaults and 'id' in defaults:
      defaults['name'] = defaults['id']
    if 'description' not in defaults and 'name' in defaults:
      defaults['description'] = defaults['name']
    if 'id' not in defaults:
      raise ValueError("Attempted to construct an object without some way of "
                       "assigning an ID.")
    super().__init__(id=defaults['id'])
    for k in defaults:
      setattr(self, k, defaults[k])
    self.properties = self.properties or set()

  def noun_phrase(self):
    if self.generic:
      # TODO: Special-case plurals here?
      return self.a()
    else:
      return self.the()

  def __str__(self):
    if hasattr(self, 'name'):
      return str(self.name)
    else:
      return str(self.id)

  def __repr__(self):
    if hasattr(self, 'name'):
      return "<something called '{}'>".format(self.name)
    else:
      return "<something with ID '{}'>".format(self.id)

  def __getattr__(self, attribute):
    if attribute.startswith(PROPERTY_ACCESSOR):
      property = attribute[len(PROPERTY_ACCESSOR):]
      return self.has_property(property)
    return self.__getattribute__(attribute)

  def get_defaults(self):
    '''
    This method returns a (deep) copy of the defaults for this object, updated
    with the defaults for all of its parents.
    '''
    defs = {}
    for parent in self.__class__.__mro__[-1:0:-1]:
      if hasattr(parent, "defaults"):
        defs.update(copy.deepcopy(getattr(parent, "defaults")))
    defs.update(copy.deepcopy(self.defaults))
    return defs

  def has_property(self, property):
    '''
    Returns True if this object's 'properties' set contains the given property.
    '''
    return property in self.properties

  # TODO: Other forms of pronouns (Pronoun objects...)?
  def Pronoun_for(self, case="subjective", person="third"):
    return self.pronoun_for(case, person).capitalize()

  def pronoun_for(self, case="objective", person="third"):
    return formatter.pronoun_for(self, case, person)

  def A(self): return self.article.title() + self.name
  def a(self): return self.article + self.name

  An = A
  an = a

  def The(self): return self.determiner.title() + self.name
  def the(self): return self.determiner + self.name

  def At(self): return self.preposition.title() + self.name
  def at(self): return self.preposition + self.name

  def At_the(self): return self.preposition.title() + self.determiner +self.name
  def at_the(self): return self.preposition + self.determiner + self.name

  def I(self): return self.pronoun_for(case="subjective", person="first")
  def me(self): return self.pronoun_for(case="objective", person="first")
  def myself(self): return self.pronoun_for(case="reflexive", person="first")
  def mine(self): return self.pronoun_for(case="possessive", person="first")
  def my(self): return self.pronoun_for(case="determinant", person="first")

  def thou(self): return self.pronoun_for(case="subjective", person="second")
  def thee(self): return self.pronoun_for(case="objective", person="second")
  def thyself(self): return self.pronoun_for(case="reflexive", person="second")
  def thine(self): return self.pronoun_for(case="possessive", person="second")
  def thy(self): return self.pronoun_for(case="determinant", person="second")

  def ey(self): return self.pronoun_for(case="subjective", person="third")
  def em(self): return self.pronoun_for(case="objective", person="third")
  def eirself(self): return self.pronoun_for(case="reflexive", person="third")
  def eirs(self): return self.pronoun_for(case="possessive", person="third")
  def eir(self): return self.pronoun_for(case="determinant", person="third")

  def capitalize(self): return self.name.capitalize()

  def title(self): return formatter.titlecase(self.name)
  def a_title(self): return formatter.titlecase(self.a())
  def the_title(self): return formatter.titlecase(self.the())

  def describe(self, indent=2):
    if self.description:
      return formatter.paragraph(
        formatter.sentence(self.description.strip()),
        indent=indent,
      )
    else:
      return ""

class Direction(Anything):
  defaults = {
    'abbreviations': set(),
    'opposite': None,
    'determiner': "",
    'article': "",
    'preposition': "",
  }

class Place(Anything):
  defaults = {
    'order': 0,
    'connections': {},
  }

class Time(Anything):
  defaults = {
    'order': 0,
  }

class Thing(Anything):
  defaults = {
    'plural': False,
    'starting_place': None,
  }

class Entity(Thing):
  defaults = {
    'gender': 'neutral',
  }

class Object(Thing):
  defaults = {}

class Event(Anything):
  defaults = {
    'order': 0,
    'verb': 'do',
    'bindings': {},
    'conditions': (),
    'description': "Something happens.",
    'occurred': False,
  }

  def check_conditions(self, state, place):
    '''
    Checks event preconditions for the given state and location.
    '''
    env = LIB(state, place)
    for cond in self.conditions:
      cond = subids(cond, state.text.all_ids(), env)
      try:
        debug.debug("Conditions", "Testing condition:", cond)
        if not eval(cond, globals(), env):
          debug.debug("Conditions", "Condition failed:", cond)
          return False
      except:
        debug.debug("Conditions", "Condition crashed:", cond)
        if debug.domainactive("Conditions"):
          sys.excepthook(*sys.exc_info())
        return False
    debug.debug("Conditions", "Conditions cleared!")
    return True

  def bind(self, state, place):
    '''
    Given a state and place, searches for and binds things from that state
    according to the bind list for this event. Returns None on failure, and a
    binding dictionary on success.
    '''
    bound = {}
    # TODO: Backtracking search for interdependent preconditions?
    for key in self.bindings:
      env = LIB(state, place)
      typ, name = key.split('/')
      debug.debug("Binding", "Searching for '{}' of type {}.".format(name, typ))
      debug.indent("Binding")
      if typ == "Time":
        targets = state.text.times.keys()
      elif typ == "Place":
        targets = state.text.places.keys()
      elif typ == "Thing":
        targets = state.text.things.keys()
      elif typ == "Entity":
        targets = [k for k, v in state.text.things.items()
                   if isinstance(v, Entity)]
      elif typ == "Object":
        targets = [k for k, v in state.text.things.items()
                   if isinstance(v, Object)]
      elif typ == "Event":
        targets = state.text.events.keys()
      elif typ == "Activity":
        targets = [k for k, v in state.text.events.items()
                   if isinstance(v, Activity)]
      elif typ == "Move":
        targets = [k for k, v in state.text.events.items()
                   if isinstance(v, Move)]
      elif typ == "DefaultMove":
        targets = [k for k, v in state.text.events.items()
                   if isinstance(v, DefaultMove)]
      elif typ == "PlayerMove":
        targets = [k for k, v in state.text.events.items()
                   if isinstance(v, PlayerMove)]
      elif typ == "ForcedMove":
        targets = [k for k, v in state.text.events.items()
                   if isinstance(v, ForcedMove)]
      elif typ == "Occurrence":
        targets = []
        for time in filter(
          lambda t: t.order <= state.time.order,
          sorted(state.text.times.values(), key=lambda t: t.order)
        ):
          targets.extend([occ.id for occ in state.trace[time]])
      elif typ == "Condition":
        targets = state.text.conditions.keys()
      else:
        targets = state.text.all_ids()
      for id in targets:
        target = state.lookup(id)
        debug.debug("Binding", "Testing target id:", id)
        locals = dict(env)
        locals.update(bound)
        locals['this'] = target
        failed = False
        for cond in self.bindings[key]:
          cond = subids(cond, state.text.all_ids(), locals)
          try:
            if not eval(cond, globals(), locals):
              failed = True
              debug.debug("Binding", "Condition failed:", cond)
              break
          except:
            debug.debug("Binding", "Condition crashed:", cond)
            if debug.domainactive("Binding"):
              sys.excepthook(*sys.exc_info())
            failed = True
            break
        if not failed:
          debug.debug("Binding", "Conditions cleared, binding target!")
          bound[name] = target
          break
      debug.exdent("Binding")
      if name not in bound or not bound[name]:
        debug.debug("Binding", "No binding established, bind step failed.")
        return None
    return bound

  def occur(self, state, place):
    '''
    Changes the given state to reflect the occurrence of this Event at the
    given Place, and returns an Occurrence specifying the parameters used. If
    the event isn't applicable, it returns None, and the state is not changed.
    '''
    debug.debug("Conditions","Checking conditions for event '{}'.".format(self))
    debug.indent("Conditions")
    if not self.check_conditions(state, place):
      debug.exdent("Conditions")
      debug.debug("Conditions", "Condition check failed.")
      return None
    debug.exdent("Conditions")
    debug.debug("Conditions", "Condition checks passed.")
    debug.debug("Binding", "Binding variables for event '{}'.".format(self))
    debug.indent("Binding")
    vars = self.bind(state, place)
    debug.exdent("Binding")
    if vars == None:
      debug.debug("Binding", "Could not bind variables.")
      return None
    self.bound = vars
    terms = LIB(state, place)
    terms.update(vars)
    if self.verb == "move":
      if not 'subject' in vars or not 'destination' in vars:
        debug.debug("Conditions", "Move event missing referent.")
        return None
      if not isinstance(vars['subject'], Thing):
        debug.debug("Conditions", "Move event with bad subject.")
        return None
      if not isinstance(vars['destination'], Place):
        debug.debug("Conditions", "Move event with bad destination.")
        return None
      state.locations[vars['subject']] = vars['destination']
    elif self.verb == "set":
      # TODO: First, implement time-varying properties!
      # TODO: ... then implement property setting via events...
      pass
    elif self.verb == "TODO: Implement more Event types!":
      pass
    self.occurred = True
    properties = dict(self.bound)
    if 'name' in properties: properties['_name'] = properties['name']
    properties['name'] = formatter.sp_format(
      "{name} happens {at time} {at the place}",
      name = self.name,
      time = state.time,
      place = place,
    ),
    if 'event' in properties: properties['_event'] = properties['event']
    properties['event'] = self
    if 'time' in properties: properties['_time'] = properties['time']
    properties['time'] = state.time
    if 'place' in properties: properties['_place'] = properties['place']
    properties['place'] = place
    if 'description' in properties:
      properties['_description'] = properties['description']
    properties['description'] = state.text.format(self.description, **terms)
    return Occurrence(**properties)

class Activity(Event):
  defaults = {
    'order': 1,
  }

class Move(Event):
  defaults = {
    'verb': "move",
    'subject': None,
    'when': None,
    'destination': None,
    'description': "{the subject} {if isinstance(subject, Entity)}{/goes/go/} to{else}{/is/are/} moved to{end} {the destination}"
  }
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    if not self.bindings and (self.subject and self.when and self.destination):
      if isinstance(self.subject, Anything):
        self.subject = self.subject.id
      if isinstance(self.when, Anything):
        self.when = self.when.id
      if isinstance(self.destination, Anything):
        self.destination = self.destination.id
      self.bindings = {
        'Thing/subject': ('this is ' + self.subject,),
        'Time/when': ('this is now is ' + self.when,),
        'Place/destination': ('this is ' + self.destination,),
      }

class DefaultMove(Move):
  defaults = { 'order': -1.5 }

class PlayerMove(Move):
  defaults = { 'order': -1 }

class ForcedMove(Move):
  defaults = { 'order': -0.5 }

class Occurrence(Anything):
  defaults = {
    'event': None,
    'time': None,
    'place': None,
    'description': "Something happens.",
  }

class Condition(Anything):
  defaults = {
    'determiner': "",
    'bindings': {},
    'requires': None
  }

# The non-things:
NoDirection = Direction(name="no direction")
Nowhere = Place(name="nowhere", determiner="", preposition="")
Never = Time(name="never", determiner="")
Nothing = Thing(name="nothing", determiner="")
Nonevent = Event(name="non-event", generic=True)
ThisIsNotACondition = Condition(name="this is not a condition", holds=False)

class TEXT(Anything):
  '''
  A Temporal EXTension is a description of some set of locations across an
  extended temporal period. It includes a list of places and a list of times,
  and by default represents each location at each time. It may also include
  things (such as people, animals, or objects), events (things that happen
  under certain conditions), and conditions (descriptions of complicated world
  state that may or may not hold at certain wherens).
  '''
  defaults = {
    'title': "The Story",
    'author': "Anonymous",
    'date': "Unknown",
    'introduction': None,
    'directions': {},
    'places': {},
    'times': {},
    'things': {},
    'events': {},
    'conditions': {},
    'actions': {},
    'aliases': multidict.multidict(),
    'start': None,
    'starting_place': None,
    'starting_time': None,
    'idfmt': "{} by {} ({})",
    'idargs': ('title', 'author', 'date'),
    'player_place': None,
    'player_time': None,
    'states': None,
  }

  def __init__(self, **kwargs):
    super().__init__(**kwargs)

    self.starting_place = self._place(self.starting_place) \
                       or (self.places and \
                           self._place(first(self.places.values()))) \
                       or Nowhere
    self.starting_time = self._time(self.starting_time) \
                       or (self.times and \
                           self._time(first(self.times.values()))) \
                       or Never
  
  @property
  def start(self):
    return (self._place(self.starting_place), self._time(self.starting_time))

  @start.setter
  def start(self, value):
    try:
      sp, st = value
    except:
      return
    self.starting_place = sp
    self.starting_time = st

  @start.deleter
  def start(self):
    self.starting_place = self._place(first(self.places.values())) or Nowhere
    self.starting_time = self._time(first(self.times.values())) or Never

  def __str__(self):
    return self.title + " by " + self.author + " (" + self.date + ")"

  def __repr__(self):
    return "<{} with {} places, {} times, {} things, and {} events>".format(
      str(self),
      len(self.places),
      len(self.times),
      len(self.things),
      len(self.events),
    )

  def init(self):
    '''
    Starts the TEXT, calling any setup code.
    '''
    self.player_place, self.player_time = self.start
    aliased = self.aliases.values()
    for id in self.all_ids():
      obj = self.lookup(id)
      if obj not in aliased:
        self.alias(obj)

  def introduce(self, stats=False):
    '''
    Returns introductory text.
    '''
    title = str(self)
    if stats:
      stats = ("{tm} times, {pl} places, {en} entities, {ob} objects, and {ev} "
               "events.\n").format(
        tm = len(self.times),
        pl = len(self.places),
        en = len(list(filter(
          lambda t: isinstance(t, Entity),
          self.things.values()
        ))),
        ob = len(list(filter(
          lambda t: isinstance(t, Object),
          self.things.values()
        ))),
        ev = len(self.events),
      )
    else:
      stats = ""
    return title + '\n' + stats + formatter.paragraph(self.introduction)

  def describe(self, where=None, when=None):
    '''
    Returns text that describes the given place at the given time. Where and
    when should be strings that name a place and a time, respectively.
    '''
    if not self.states:
      debug.debug("Simulation", "Initial simulation.")
      self.simulate()
      assert(self.states)
    where = self._place(where or self.player_place)
    when = self._time(when or self.player_time)
    state = self.states[when]
    things = [thing for thing in state.locations
              if state.locations[thing] == where]
    if things:
      stuff = formatter.sentence(formatter.sp_format(
        "{things:list} {/is/are/} {if number_of(things) >2}all {end}here.",
        things = things,
      ))
    else:
      stuff = ''
    return """\
{place}, {time}
{place_desc}{stuff}{events}\
""".format(
      place = where.title(),
      time = when.title(),
      place_desc = where.describe(),
      stuff = formatter.paragraph(stuff, indent=-2),
      events = ''.join(occurrence.describe(indent=-2)
                       for occurrence in state.trace[when]
                       if occurrence.place == where),
    )

  def handle_action(self, input):
    '''
    Handles the given input according to the action list for the TEXT.
    '''
    vars = LIB(self.states[self.player_time], self.player_place)
    for a in self.actions:
      # TODO: Make this more elegant?
      if self.actions[a].handle(
        input,
        (self.states[self.player_time], self.player_place),
        globals(),
        locals=vars
      ):
        return
    close = sorted(
      filter(lambda a: a.score > 0, self.actions.values()),
      key=lambda a: a.score,
    )
    if close:
      closest = close[-1]
      print("""\
Sorry, it seemed like you were trying to use the command '{action}', but I
didn't understand your command. Try 'help {action}' if you are confused.
""".format(action=closest.id))
    else:
      print("Sorry, I didn't understand that at all.")

  def out_of_bounds(self, where, when):
    '''
    Tests whether a time/location is out-of-bounds. Uses the outofbounds
    property.
    '''
    # TODO: Make this richer.
    return where.has_property_outofbounds

  # Action functions:

  def move_player(self, first=None, second=None):
    '''
    Moves the player. Takes two arguments, which each may be a place or time.
    If they are both either places or times, the second is ignored. If only one
    is given, the other type (place/time) defaults to the current value for the
    player.
    '''
    if not first:
      return False
    where = self._place(first) or self._place(second) or self.player_place
    when = self._time(first) or self._time(second) or self.player_time
    if not self.out_of_bounds(where, when):
      self.player_place, self.player_time = where, when
    else:
      # TODO: Report this in a better fashion?
      output("Sorry, but that's out-of-bounds.")
      return False
    return True

  def move_thing(self, thing, first=None, second=None):
    # TODO: Make this generic (check logic is duplicated in
    # Event.happens/execute?)
    '''
    Moves a Thing to the given place during the given time. If no time is
    specified, the current time of the player is used. If no place is
    specified, nothing happens. Returns a string describing the error
    condition, or None to indicate success.
    '''
    thing = self._thing(thing)
    if not thing:
      return "the target was invalid"
    where = self._place(first) or self._place(second)
    if not where:
      return "the destination was invalid"
    when = self._time(first) or self._time(second) or self.player_time
    ename = "player moves " + thing.name + " " + when.at()
    event = self._event(ename)
    if event:
      event.bindings['Place/destination'] = ('this is ' + where.id,)
      debug.debug("Simulation", "Simulating for object movement.")
      self.simulate()
    else:
      self.add(PlayerMove(
        name=ename,
        subject=thing,
        when=when,
        destination=where,
      ))
      debug.debug("Simulation", "Simulating for object movement.")
      self.simulate()
      if not self._event(ename).occurred:
        self.release_thing(thing, when)
        return "the action was impossible"
    loc = self.states[when].locations[thing]
    if loc != where:
      return "the target moved to {} instead".format(loc.the())

  def release_thing(self, thing, when=None):
    '''
    Removes any player move events for the given object at the given time. This
    causes the object to revert back to normal inertia.
    '''
    thing = self._thing(thing)
    if not thing:
      return
    when = self._time(when) or self.player_time
    ename = "player moves " + thing.name + " " + when.at()
    event = self._event(ename)
    if event:
      self.remove(event)
    debug.debug("Simulation", "Simulating for object release.")
    self.simulate()

  def add(self, anything):
    if isinstance(anything, Direction):
      self.directions[anything.id] = anything
    if isinstance(anything, Place):
      self.places[anything.id] = anything
    if isinstance(anything, Time):
      self.times[anything.id] = anything
    if isinstance(anything, Thing):
      self.things[anything.id] = anything
    if isinstance(anything, Event):
      self.events[anything.id] = anything
    if isinstance(anything, Condition):
      self.conditions[anything.id] = anything
    self.alias(anything)

  def alias(self, anything):
    names = set(anything.name.split())
    if hasattr(anything, "abbreviations") and anything.abbreviations:
      for abr in anything.abbreviations:
        names.add(abr)
    names.add(anything.name)
    for n in names:
      if n.lower() not in names:
        names.add(n.lower())
    for s in formatter.STOPWORDS:
      if s in names:
        names.remove(s)
      if s.capitalize() in names:
        names.remove(s.capitalize())
    for n in names:
      self.aliases.map(n, anything)

  def remove(self, anything):
    if isinstance(anything, Place) and anything in self.places:
      del self.places[anything.id]
    if isinstance(anything, Time) and anything in self.times:
      del self.times[anything.id]
    if isinstance(anything, Thing) and anything in self.things:
      del self.things[anything.id]
    elif isinstance(anything, Event) and anything in self.events:
      del self.events[anything.id]
    if isinstance(anything, Condition) and anything in self.conditions:
      del self.conditions[anything.id]
    self.aliases.unmap_all(anything)

  def simulate(self):
    '''
    Forward-simulates the entire TEXT, filling in the surface-layer state given
    the event set. This should be called after any player action that changes
    the TEXT state (usually by adding or removing events).
    '''
    for e in self.events:
      self.events[e].occurred = False
    timesequence = sorted(self.times.values(), key=lambda t: t.order)
    self.states = {}
    last = None
    debug.debug("Simulation", "Starting...")
    debug.indent("Simulation")
    for time in timesequence:
      debug.debug("Simulation", "Time is", time)
      debug.indent("Simulation")
      if last:
        self.states[time] = State(
          name = "the state of " + self.title + " " + time.at(),
          text = self,
          time = time,
          previous = self.states[last]
        )
      else:
        self.states[time] = State(
          name = "the state of " + self.title + " " + time.at(),
          text = self,
          time = time,
          locations = {thing: (thing.starting_place or Nowhere)
                       for thing in self.things.values()},
        )
      debug.debug("Simulation", "Locations:", self.states[time].locations)
      allevents = sorted(self.events.values(), key=lambda e: e.order)
      allplaces = [Nowhere] + sorted(
        self.places.values(),
        key=lambda p: p.order
      )
      occurred = True
      while occurred:
        occurred = None
        # TODO: Recurring events?
        for event in filter(lambda e: not e.occurred, allevents):
          debug.debug("Simulation", "Event:", event)
          debug.indent("Simulation")
          for place in allplaces:
            debug.debug("Simulation", "Place:", place)
            debug.indent("Simulation")
            occurred = self.states[time].occur(event, place) or occurred
            if occurred:
              debug.debug("Simulation", "Event occurred!")
              debug.exdent("Simulation")
              break
            debug.exdent("Simulation")
          if occurred:
            debug.exdent("Simulation")
            break
          debug.exdent("Simulation")

      last = time
      debug.exdent("Simulation")
    debug.exdent("Simulation")
    debug.debug("Simulation", " ...done.")

  def _place(self, where):
    '''
    Looks up a place, returning None if the given place isn't found.
    '''
    if isinstance(where, Place):
      return where
    if where in self.places:
      return self.places[where]
    return None

  def _time(self, when):
    '''
    Like _place, but looks up a time.
    '''
    if isinstance(when, Time):
      return when
    if when in self.times:
      return self.times[when]
    return None

  def _thing(self, which):
    '''
    Like _place, but looks up a thing.
    '''
    if isinstance(which, Thing):
      return which
    if which in self.things:
      return self.things[which]
    return None

  def _event(self, which):
    '''
    Like _place, but looks up an event.
    '''
    if isinstance(which, Event):
      return which
    if which in self.events:
      return self.events[which]
    return None

  def _condition(self, which):
    '''
    Like _place, but looks up an event.
    '''
    if isinstance(which, Condition):
      return which
    if which in self.conditions:
      return self.conditions[which]
    return None

  def all_ids(self):
    '''
    Returns a set of all object IDs in this TEXT (not including the ID of the
    text itself).
    '''
    result = set()
    for valset in (
      self.places,
      self.directions,
      self.times,
      self.things,
      self.events,
      self.conditions,
      self.actions
    ):
      result = result.union(set(valset.keys()))
    return result

  def lookup(self, id):
    '''
    Returns the object corresponding to the given id, or None.
    '''
    for valset in (
      self.places,
      self.directions,
      self.times,
      self.things,
      self.events,
      self.conditions,
      self.actions
    ):
      if id in valset:
        return valset[id]
    return None

  def lookup_name(self, name):
    '''
    Looks up a name within the text, using the aliases list. Returns None if
    no such name exists, or a set of results if that name is mapped to
    something.
    '''
    if name in self.aliases:
      return self.aliases[name]
    else:
      return None

  def format(self, content, *args, **kwargs):
    '''
    Calls the formatter.sp_format method with an appropriate _CODE_PREP_
    argument.
    '''
    kwargs.update({"_CODE_PREP_": lambda c: subids(c, self.all_ids(), kwargs)})
    return formatter.sp_format(content, *args, **kwargs)


class State(Anything):
  '''
  A State represents the state of the world in a TEXT at a particular Time.
  Whereas an Event happens within a TEXT, events change one State into another.
  States have a text, a time, a mapping from Things to Places called locations,
  and a dictionary of lists of events called the trace.
  '''
  defaults = {
    'text': None,
    'time': Never,
    'previous': None,
    'locations': {},
    'trace': {},
    'occurrences': {},
  }
  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    if self.previous:
      self.locations = dict(self.previous.locations)
      for t in self.previous.trace:
        self.trace[t] = self.previous.trace[t][:]
      self.occurrences = dict(self.previous.occurrences)
    self.trace[self.time] = []

  def occur(self, event, where):
    '''
    Causes the given event to occur at the given place (and at the current time
    for this State). The event is added to the trace for this state, and any
    operations that it specifies are carried out. The Event's preconditions are
    checked beforehand, and if not met, nothing happens (including the Event),
    and this method returns False. If the event fires, True is returned.
    '''
    occurrence = event.occur(self, where)
    if not occurrence:
      return False
    self.trace[self.time].append(occurrence)
    self.occurrences[occurrence.id] = occurrence
    return True

  def lookup(self, id):
    '''
    Tries lookup on the text, and then looks for an Occurrence if the ID
    doesn't match anything in the TEXT.
    '''
    result = self.text.lookup(id)
    if result:
      return result
    elif id in self.occurrences:
      return self.occurrences[id]
